@@ -1,14 +1,14 @@
-from gscraper.base import UPLOAD_CONTEXT
-from gscraper.base.session import TypedDict, TypedRecords, BaseSession
-from gscraper.base.parser import Schema, Field, INVALID_OBJECT_MSG, INVALID_OBJECT_TYPE_MSG
-from gscraper.base.types import _KT, TypeHint, LogLevel, IndexLabel, Datetime, RenameMap
+from gscraper.base.abstract import TypedDict, TypedRecords, GCLOUD_CONTEXT
+from gscraper.base.session import BaseSession, Schema, Field, INVALID_OBJECT_MSG, INVALID_OBJECT_TYPE_MSG
+
+from gscraper.base.types import _KT, TypeHint, IndexLabel, RenameMap
 from gscraper.base.types import TabularData, Account, PostData, is_records, from_literal
 
 from gscraper.utils.cast import cast_str, cast_list, cast_float, cast_int, cast_datetime_format
 from gscraper.utils.date import get_datetime, get_timestamp, get_time, get_date, DATE_UNIT
 from gscraper.utils.logs import log_table
-from gscraper.utils.map import df_exists, df_empty, data_exists
-from gscraper.utils.map import iloc, to_array, kloc, drop_dict, cloc, to_dataframe, apply_df, apply_data
+from gscraper.utils.map import isna, df_exists, df_empty, to_array, get_scala
+from gscraper.utils.map import kloc, to_dict, to_records, cloc, to_dataframe, apply_df, apply_data
 
 from google.oauth2 import service_account
 from google.oauth2.service_account import IDTokenCredentials
@@ -37,16 +37,9 @@ ENV_PATH = "env/"
 GCLOUD_ACCOUNT = ENV_PATH+"gcloud.json"
 GCLOUD_DATA = ENV_PATH+"data.json"
 
-KEY = "key"
-SHEET = "sheet"
-FIELDS = "fields"
-
-TABLE = "table"
-QUERY = "query"
-PID = "project_id"
-
-MODE = "mode"
-DATA = "data"
+KEY, SHEET, FIELDS = "key", "sheet", "fields"
+TABLE, QUERY, PID = "table", "query", "project_id"
+MODE, DATA = "mode", "data"
 
 FROM_GS, TO_GBQ = ["from_key", "from_sheet"], ["to_table", "to_pid"]
 FROM_GBQ, TO_GS = ["from_query", "from_pid"], ["to_key", "to_sheet"]
@@ -59,12 +52,14 @@ UPLOAD = lambda name=str(): f"upload_{name}" if name else "upload"
 ############################# Messages ############################
 ###################################################################
 
+INVALID_QUERY_MSG = "To update data, parameters for source and destination are required."
+
 INVALID_AXIS_MSG = lambda axis: f"'{axis}' is not valid axis. Only allowed in table(-1), each column(0), each row(1)."
 
 INVALID_GS_ACTION_MSG = lambda action: f"'{action}' is not valid action for gspread task."
 
-INVALID_SCHEMA_MSG = "Please verify that the structure and data types in the DataFrame match the schema of the destination table."
-INVALID_UPSERT_KEY_MSG = "Please verify that a primary key exists and is in both DataFrame objects."
+SCHEMA_MISMATCH_MSG = "Please verify that the structure and data types in the DataFrame match the schema of the destination table."
+INVALID_PRIMARY_KEY_MSG = "Please verify that a primary key exists and is in both DataFrame objects."
 
 BIGQUERY_PARTITION_MSG = "Uploading partitioned data to Google BigQuery"
 
@@ -157,9 +152,13 @@ class BigQuerySchema(Schema):
     def __init__(self, *args: BigQueryField):
         TypedRecords.__init__(self, *[validate_gbq_field(field) for field in args])
 
+    def get_primary_key(self, index=0) -> str:
+        keys = [field.get("name") for field in self if isinstance(field, Dict) and field.get("mode") == "REQUIRED"]
+        return get_scala(keys, index=index)
 
-def validate_gbq_schema(schema: Any) -> BigQuerySchema:
-    if not schema: return
+
+def validate_gbq_schema(schema: Any, optional=False) -> BigQuerySchema:
+    if (not schema) and optional: return
     elif isinstance(schema, BigQuerySchema): return schema
     elif isinstance(schema, List): return BigQuerySchema(*schema)
     else: raise TypeError(INVALID_OBJECT_TYPE_MSG(schema, BIGQUERY_SCHEMA))
@@ -172,17 +171,26 @@ def validate_gbq_schema(schema: Any) -> BigQuerySchema:
 NumericiseIgnore = Union[Sequence[int], bool]
 
 class GspreadReadContext(TypedDict):
-    def __init__(self, key: str, sheet: str, fields: IndexLabel, axis=0, dropna=True, strict=True,
-                unique=False, default: Optional[Any]=None, head=1, headers: Optional[IndexLabel]=None,
-                str_cols: Optional[NumericiseIgnore]=None, arr_cols: Optional[IndexLabel]=None,
-                to: Optional[Literal["desc","name"]]=None, rename: Optional[RenameMap]=None):
+    def __init__(self, key: str, sheet: str, fields: Optional[IndexLabel]=None, default: Optional[Any]=None,
+                if_null: Literal["drop","pass"]="pass", head=1, headers: Optional[IndexLabel]=None,
+                str_cols: Optional[NumericiseIgnore]=None, to: Optional[Literal["desc","name"]]="name",
+                return_type: Optional[TypeHint]="dataframe", rename: Optional[RenameMap]=None):
         super().__init__(key=key, sheet=sheet, fields=fields)
-        self.update_default(dict(axis=0, dropna=True, strict=True, unique=False, head=1),
-            axis=axis, dropna=dropna, strict=strict, unique=unique, default=default, head=head,
-            headers=headers, str_cols=str_cols, arr_cols=arr_cols, to=to, rename=rename)
+        self.update_default(dict(if_null="pass", head=1, to="name", return_type="dataframe"),
+            default=default, if_null=if_null, head=head, headers=headers,
+            str_cols=str_cols, to=to, return_type=return_type, rename=rename)
 
 
-GoogleQueryContext = GspreadReadContext
+class GoogleQueryContext(GspreadReadContext):
+    def __init__(self, key: str, sheet: str, fields: IndexLabel, default: Optional[Any]=None,
+                if_null: Literal["drop","pass"]="drop", axis=0, dropna=True, strict=True, unique=False,
+                head=1, headers: Optional[IndexLabel]=None, str_cols: Optional[NumericiseIgnore]=None,
+                arr_cols: Optional[IndexLabel]=None, to: Optional[Literal["desc","name"]]="name",
+                return_type: Optional[TypeHint]="dataframe", rename: Optional[RenameMap]=None):
+        super().__init__(key, sheet, fields, default, if_null, head, headers, str_cols, to, return_type, rename)
+        self.update_default(dict(axis=0, dropna=True, strict=True, unique=False),
+            axis=axis, dropna=dropna, strict=strict, unique=unique, arr_cols=arr_cols)
+
 
 class GoogleQueryInfo(TypedDict):
     def __init__(self, **context: GoogleQueryContext):
@@ -191,36 +199,34 @@ class GoogleQueryInfo(TypedDict):
 
 class GoogleQueryReader(BaseSession):
     __metaclass__ = ABCMeta
-    operation = "gcloudQuery"
-
-    def __init__(self, queryInfo: GoogleQueryInfo=dict(), account: Account=dict(),
-                logName=str(), logLevel: LogLevel="WARN", logFile=str(),
-                debug: List[str]=list(), extraSave: List[str]=list(), interrupt=str(), localSave=False, **context):
-        BaseSession.__init__(self, **self.from_locals(locals(), drop=["queryInfo","account"]))
-        self.set_query(queryInfo, account)
+    operation = "googleQueryReader"
 
     def set_query(self, queryInfo: GoogleQueryInfo=dict(), account: Account=dict()):
         for name, queryContext in queryInfo.items():
-            if len(kloc(queryContext, [KEY, SHEET, FIELDS], if_null="drop")) == 3:
-                self.set_gs_query(**queryContext, name=name, account=account)
+            if not isinstance(queryInfo, Dict): continue
+            elif len(kloc(queryContext, [KEY, SHEET, FIELDS], if_null="drop")) != 3: continue
+            elif "if_null" not in queryContext: queryContext["if_null"] = "drop"
+            data = self.read_gspread(**queryContext, name=name, account=account)
+            self.update(self.get_values_by_axis(to_dataframe(data), **queryContext))
 
-    def set_gs_query(self, key: str, sheet: str, fields: IndexLabel, axis=0, dropna=True, strict=True, unique=False,
-                    default=None, head=1, headers=None, str_cols: NumericiseIgnore=list(),
-                    arr_cols: IndexLabel=list(), to: Optional[Literal["desc","name"]]=None,
-                    rename: RenameMap=dict(), name=str(), account: Account=dict()):
-        context = dict(default=default, head=head, headers=headers, numericise_ignore=str_cols,
-                        return_type="dataframe", rename=self.get_rename_map(to=to, renameMap=rename))
-        data = read_gspread(key, sheet, account, fields=cast_list(fields), if_null="drop", **context)
-        self.checkpoint(READ(name), where="set_gs_query", msg={KEY:key, SHEET:sheet, DATA:data}, save=data)
+    def read_gspread(self, key: str, sheet: str, fields: IndexLabel=list(), default=None,
+                    if_null: Literal["drop","pass"]="pass", head=1, headers=None,
+                    str_cols: NumericiseIgnore=list(), to: Optional[Literal["desc","name"]]="name",
+                    return_type: Optional[TypeHint]="dataframe", rename: Optional[RenameMap]=None,
+                    name=str(), account: Account=dict()) -> TabularData:
+        context = dict(default=default, if_null=if_null, head=head, headers=headers, numericise_ignore=str_cols,
+                        return_type=return_type, rename=(rename if rename else self.get_rename_map(to=to)))
+        data = read_gspread(key, sheet, account=account, fields=fields, **context)
+        self.checkpoint(READ(name), where="read_gspread", msg={KEY:key, SHEET:sheet, DATA:data}, save=data)
         self.logger.info(log_table(data, name=name, key=key, sheet=sheet, dump=self.logJson))
-        self.update(self.by_axis(data, axis, dropna, strict, unique, arr_cols))
+        return data
 
-    def by_axis(self, df: pd.DataFrame, axis=0, dropna=True, strict=True, unique=False,
-                arr_cols: IndexLabel=list()) -> Dict[_KT,Union[List,Any]]:
+    def get_values_by_axis(self, df: pd.DataFrame, axis=0, dropna=True, strict=True, unique=False,
+                            arr_cols: IndexLabel=list(), **context) -> Dict[_KT,Union[List,Any]]:
         if axis not in (-1,0,1): raise ValueError(INVALID_AXIS_MSG(axis))
-        elif axis == -1: return df.to_dict("list")
+        elif axis == -1: return to_dict(df, "list", depth=2)
         arr_cols = cast_list(arr_cols)
-        data = (df.T if axis == 1 else df).to_dict("list")
+        data = to_dict((df.T if axis == 1 else df), "list", depth=2)
         for __key, __values in data.copy().items():
             values = to_array(__values, dropna=dropna, strict=strict, unique=unique)
             if (len(values) < 2) and (__key not in arr_cols):
@@ -237,9 +243,9 @@ class GspreadUpdateContext(TypedDict):
     def __init__(self, key: str, sheet: str, mode: Literal["replace","append","upsert"]="append",
                 cell: Optional[str]=None, base_sheet: Optional[str]=None, default: Optional[Any]=None,
                 head=1, headers: Optional[IndexLabel]=None, str_cols: Optional[NumericiseIgnore]=None,
-                to: Optional[Literal["desc","name"]]=None, rename: Optional[RenameMap]=None):
+                to: Optional[Literal["desc","name"]]="desc", rename: Optional[RenameMap]=None):
         super().__init__(key=key, sheet=sheet)
-        self.update_default(dict(mode="append", head=1),
+        self.update_default(dict(mode="append", head=1, to="desc"),
             mode=mode, cell=cell, base_sheet=base_sheet, default=default,
             head=head, headers=headers, str_cols=str_cols, to=to, rename=rename)
 
@@ -248,34 +254,32 @@ class BigQueryContext(TypedDict):
     def __init__(self, table: str, project_id: str, mode: Literal["fail","replace","append","upsert"]="append",
                 base_query: Optional[str]=None, schema: Optional[BigQuerySchema]=None,
                 progress=True, partition: Optional[str]=None,
-                partition_by: Literal["auto","second","minute","hour","day","month","year","date"]="auto"):
+                partition_by: Optional[Literal["auto","second","minute","hour","day","month","year","date"]]="auto"):
         super().__init__(table=table, project_id=project_id)
         self.update_default(dict(mode="append", progress=True, partition_by="auto"),
-            mode=mode, base_query=base_query, schema=validate_gbq_schema(schema), progress=progress,
+            mode=mode, base_query=base_query, schema=validate_gbq_schema(schema, optional=True), progress=progress,
             partition=partition, partition_by=partition_by)
 
 
-class GsToGbqContext(TypedDict):
-    def __init__(self, from_key: str, from_sheet: str, to_table: str, to_pid: str,
-                mode: Literal["fail","replace","append","upsert"]="append", schema: Optional[BigQuerySchema]=None,
-                progress=True, default: Optional[Any]=None, head=1,
-                headers: Optional[IndexLabel]=None, str_cols: Optional[NumericiseIgnore]=None,
-                to: Optional[Literal["desc","name"]]=None, rename: Optional[RenameMap]=None):
-        super().__init__(from_key=from_key, from_sheet=from_sheet, to_table=to_table, to_pid=to_pid)
-        self.update_default(dict(mode="append", progress=True, head=1),
-            default=default, head=head, headers=headers, str_cols=str_cols, to=to, rename=rename,
-            mode=mode, schema=validate_gbq_schema(schema), progress=progress)
-
-
-class GbqToGsContext(TypedDict):
-    def __init__(self, from_query: str, from_pid: str, to_key: str, to_sheet: str,
-                mode: Literal["fail","replace","append","upsert"]="append", cell: Optional[str]=None):
-        super().__init__(from_query=from_query, from_pid=from_pid, to_key=to_key, to_sheet=to_sheet)
-        self.update_default(dict(mode="append"), mode=mode, cell=cell)
+class GoogleUpdateContext(TypedDict):
+    def __init__(self, from_key=str(), from_sheet=str(), from_query=str(), from_pid=str(),
+                to_key=str(), to_sheet=str(), to_table=str(), to_pid=str(),
+                mode: Literal["fail","replace","append","upsert"]="append", default: Optional[Any]=None,
+                head=1, headers: Optional[IndexLabel]=None, str_cols: Optional[NumericiseIgnore]=None,
+                to: Optional[Literal["desc","name"]]="desc", rename: Optional[RenameMap]=None, cell: Optional[str]=None,
+                schema: Optional[BigQuerySchema]=None, progress=True, partition: Optional[str]=None,
+                partition_by: Optional[Literal["auto","second","minute","hour","day","month","year","date"]]="auto"):
+        query = kloc(locals(), FROM_GS + FROM_GBQ + TO_GS + TO_GBQ, if_null="drop")
+        if len(query) != 4: raise ValueError(INVALID_QUERY_MSG)
+        super().__init__(**query)
+        self.update_default(dict(mode="append", progress=True, head=1, to="desc"),
+            mode=mode, default=default, head=head, headers=headers, str_cols=str_cols, to=to, rename=rename,
+            cell=cell, schema=validate_gbq_schema(schema, optional=True), progress=progress,
+            partition=partition, partition_by=partition_by)
 
 
 GoogleUploadMode = Literal["fail","replace","append","upsert"]
-GoogleUploadContext = Union[GspreadUpdateContext, BigQueryContext, GsToGbqContext, GbqToGsContext]
+GoogleUploadContext = Union[GspreadUpdateContext, BigQueryContext, GoogleUpdateContext]
 
 class GoogleUploadInfo(TypedDict):
     def __init__(self, **context: GoogleUploadContext):
@@ -284,29 +288,22 @@ class GoogleUploadInfo(TypedDict):
 
 class GoogleUploader(BaseSession):
     __metaclass__ = ABCMeta
-    operation = "gcloudUploader"
+    operation = "googleUploader"
     uploadStatus = defaultdict(bool)
 
-    def __init__(self, uploadInfo: GoogleUploadInfo=dict(), reauth=False, audience=str(), account: Account=dict(),
-                logName=str(), logLevel: LogLevel="WARN", logFile=str(),
-                debug: List[str]=list(), extraSave: List[str]=list(), interrupt=str(), localSave=False, **context):
-        BaseSession.__init__(self, **self.from_locals(locals()))
-
     def upload_data(self, data: TabularData, uploadInfo: GoogleUploadInfo=dict(), reauth=False,
                     audience=str(), account: Account=dict(), credentials: Optional[IDTokenCredentials]=None, **context):
-        __exists = data_exists(data)
-        if __exists: data = to_dataframe(data).copy()
-        context = UPLOAD_CONTEXT(account=account, **context)
+        data = to_dataframe(data)
+        context = GCLOUD_CONTEXT(account=account, **context)
         gbq_auth = dict(reauth=reauth, audience=audience, credentials=credentials)
         for name, uploadContext in uploadInfo.items():
-            if __exists and (len(kloc(uploadContext, [KEY, SHEET], if_null="drop")) == 2):
+            if not isinstance(uploadContext, Dict): status = False
+            elif (not data.empty) and (len(kloc(uploadContext, [KEY, SHEET], if_null="drop")) == 2):
                 status = self.upload_gspread(data=data.copy(), **uploadContext, name=name, **context)
-            elif __exists and (len(kloc(uploadContext, [TABLE, PID], if_null="drop")) == 2):
+            elif (not data.empty) and (len(kloc(uploadContext, [TABLE, PID], if_null="drop")) == 2):
                 status = self.upload_gbq(data=data.copy(), **uploadContext, name=name, **gbq_auth, **context)
-            elif len(kloc(uploadContext, FROM_GS + TO_GBQ, if_null="drop")) == 4:
-                status = self.gspread_to_gbq(**uploadContext, name=name, **gbq_auth, **context)
-            elif len(kloc(uploadContext, FROM_GBQ + TO_GS, if_null="drop")) == 4:
-                status = self.gbq_to_gspread(**uploadContext, name=name, **gbq_auth, **context)
+            elif len(kloc(uploadContext, FROM_GS + FROM_GBQ + TO_GS + TO_GBQ, if_null="drop")) == 4:
+                status = self.update_data(**uploadContext, name=name, **gbq_auth, **context)
             else: status = False
             self.uploadStatus[name] = status
 
@@ -315,29 +312,26 @@ class GoogleUploader(BaseSession):
     ###################################################################
 
     @BaseSession.catch_exception
-    def upload_gspread(self, key: str, sheet: str, data: pd.DataFrame, mode: Literal["replace","append","upsert"]="append",
-                        cell=str(), base_sheet=str(), default=None, head=1, headers=None,
-                        str_cols: NumericiseIgnore=list(), to: Optional[Literal["desc","name"]]=None,
-                        rename: RenameMap=dict(), name=str(), account: Account=dict(), **context) -> bool:
-        data = self.map_gs_data(data, name=name, **context)
+    def upload_gspread(self, key: str, sheet: str, data: pd.DataFrame,
+                        mode: Literal["replace","append","upsert"]="append", cell=str(), base_sheet=str(), default=None,
+                        head=1, headers=None, str_cols: NumericiseIgnore=list(), to: Optional[Literal["desc","name"]]="desc",
+                        rename: Optional[RenameMap]=None, name=str(), account: Account=dict(), **context) -> bool:
         if base_sheet or (mode == "upsert"):
             base_sheet = sheet if mode == "upsert" else base_sheet
-            base = self.read_gs_base(key, base_sheet, name, account, default, head, headers, str_cols, to, rename)
+            base = self.read_gs_base(key, base_sheet, name, account, default, head, headers, str_cols, rename=rename)
             data = self.map_gs_base(data, base, name=name, **context)
+        data = self.map_gs_data(data, name=name, **context)
         self.checkpoint(UPLOAD(name), where="upload_gspread", msg={KEY:key, SHEET:sheet, MODE:mode, DATA:data}, save=data)
         self.logger.info(log_table(data, name=name, key=key, sheet=sheet, mode=mode, dump=self.logJson))
         cell, clear = ("A2" if mode == "replace" else (cell if cell else str())), (True if mode == "replace" else clear)
-        update_gspread(key, sheet, data, account, cell=cell, clear=clear)
+        update_gspread(key, sheet, data, account=account, cell=cell, clear=clear)
         return True
 
-    def map_gs_data(self, data: pd.DataFrame, name=str(), **context) -> pd.DataFrame:
-        return data
-
     def read_gs_base(self, key: str, sheet: str, name=str(), account: Account=dict(), default=None,
                     head=1, headers=None, str_cols: NumericiseIgnore=list(),
-                    to: Optional[Literal["desc","name"]]=None, rename: RenameMap=dict()) -> pd.DataFrame:
-        data = read_gspread(key, sheet, account, default=default, head=head, headers=headers,
-                            numericise_ignore=str_cols, rename=self.get_rename_map(to=to, renameMap=rename))
+                    to: Optional[Literal["desc","name"]]="name", rename: Optional[RenameMap]=None) -> pd.DataFrame:
+        data = read_gspread(key, sheet, account=account, default=default, head=head, headers=headers,
+                            numericise_ignore=str_cols, rename=(rename if rename else self.get_rename_map(to=to)))
         self.checkpoint(READ(name), where="read_gs_base", msg={KEY:key, SHEET:sheet, DATA:data}, save=data)
         self.logger.info(log_table(data, name=name, key=key, sheet=sheet, dump=self.logJson))
         return data
@@ -345,6 +339,9 @@ class GoogleUploader(BaseSession):
     def map_gs_base(self, data: pd.DataFrame, base: pd.DataFrame, name=str(), **context) -> pd.DataFrame:
         return cloc(data, base.columns, if_null="pass")
 
+    def map_gs_data(self, data: pd.DataFrame, name=str(), **context) -> pd.DataFrame:
+        return data
+
     ###################################################################
     ######################### Google BigQuery #########################
     ###################################################################
@@ -356,13 +353,13 @@ class GoogleUploader(BaseSession):
                     partition_by: Literal["auto","second","minute","hour","day","date"]="auto",
                     name=str(), reauth=False, audience=str(), account: Account=dict(),
                     credentials: Optional[IDTokenCredentials]=None, **context) -> bool:
-        schema = schema if schema and is_records(schema, how="all") else self.get_gbq_schema(name=name, schema=schema, **context)
-        data = self.map_gbq_data(data, schema=schema, name=name, **context)
+        schema = validate_gbq_schema(schema if schema else self.get_gbq_schema(name=name, schema=schema, **context))
         gbq_auth = dict(reauth=reauth, audience=audience, account=account, credentials=credentials)
         if base_query or (mode == "upsert"):
             base_query = table if mode == "upsert" else base_query
-            base = self.read_gbq_base(base_query, project_id, name, **gbq_auth)
-            data = self.map_gbq_base(data, base, name=name, schema=schema, **context)
+            base = self.read_gbq_base(base_query, project_id, name=name, **gbq_auth)
+            data = self.map_gbq_base(data, base, table=table, project_id=project_id, schema=schema, name=name, **gbq_auth, **context)
+        data = self.map_gbq_data(data, table=table, project_id=project_id, schema=schema, name=name, **gbq_auth, **context)
         self.checkpoint(UPLOAD(name), where="upload_gbq", msg={TABLE:table, PID:project_id, MODE:mode, DATA:data}, save=data)
         self.logger.info(log_table(data, name=name, table=table, pid=project_id, mode=mode, schema=schema, dump=self.logJson))
         to_gbq(table, project_id, data, if_exists=("replace" if mode == "upsert" else mode), schema=schema, progress=progress,
@@ -372,47 +369,49 @@ class GoogleUploader(BaseSession):
     def get_gbq_schema(self, name=str(), **context) -> BigQuerySchema:
         ...
 
-    def map_gbq_data(self, data: pd.DataFrame, schema: BigQuerySchema, name=str(), **context) -> pd.DataFrame:
-        columns = [field["name"] for field in schema if field["name"] in data]
-        data = cloc(data, columns, if_null="drop", reorder=True)
-        if len(data.columns) != len(columns): raise ValueError(INVALID_SCHEMA_MSG)
-        else: return data
-
     def read_gbq_base(self, query: str, project_id: str, name=str(), reauth=False, audience=str(),
                         account: Account=dict(), credentials: Optional[IDTokenCredentials]=None) -> pd.DataFrame:
-        data = read_gbq(query, project_id, reauth, audience, account, credentials)
+        data = read_gbq(query, project_id, reauth=reauth, audience=audience, account=account, credentials=credentials)
         self.checkpoint(READ(name), where="read_gbq_base", msg={QUERY:query, PID:project_id}, save=data)
         self.logger.info(log_table(data, name=name, query=query, pid=project_id, dump=self.logJson))
         return data
 
-    def map_gbq_base(self, data: pd.DataFrame, base: pd.DataFrame, key=str(),
-                    schema: Optional[BigQuerySchema]=None, **context) -> pd.DataFrame:
-        key = _validate_upsert_key(data, base, key, schema)
+    def map_gbq_base(self, data: pd.DataFrame, base: pd.DataFrame, schema: BigQuerySchema, **context) -> pd.DataFrame:
+        key = validate_primary_key(data, base, schema=schema)
         data = data.set_index(key).combine_first(base.set_index(key)).reset_index()
         return data[data[key].notna()].drop_duplicates(key)
 
+    def map_gbq_data(self, data: pd.DataFrame, schema: Optional[BigQuerySchema]=None, name=str(), **context) -> pd.DataFrame:
+        columns = [field["name"] for field in schema if field["name"] in data]
+        data = cloc(data, columns, if_null="drop", reorder=True)
+        if len(data.columns) != len(columns): raise ValueError(SCHEMA_MISMATCH_MSG)
+        else: return data
+
     ###################################################################
-    ########################### Cross Upload ##########################
+    ########################### Update Data ###########################
     ###################################################################
 
     @BaseSession.catch_exception
-    def gspread_to_gbq(self, from_key: str, from_sheet: str, to_table: str, to_pid: str,
-                        default=None, head=1, headers=None, str_cols: NumericiseIgnore=list(),
-                        mode: Literal["fail","replace","append","upsert"]="append",
-                        schema: Optional[BigQuerySchema]=None, progress=True,
-                        name=str(), reauth=False, audience=str(), account: Account=dict(),
-                        credentials: Optional[IDTokenCredentials]=None, **context) -> bool:
-        data = self.read_gs_base(from_key, from_sheet, name, account, default, head, headers, str_cols)
-        return self.upload_gbq(to_table, to_pid, data, mode, schema, progress=progress, name=name,
-                                reauth=reauth, audience=audience, account=account, credentials=credentials, **context)
-
-    @BaseSession.catch_exception
-    def gbq_to_gspread(self, from_query: str, from_pid: str, to_key: str, to_sheet: str,
-                        mode: Literal["fail","replace","append","upsert"]="append", cell=str(),
-                        name=str(), reauth=False, account: Account=dict(), audience=str(),
-                        credentials: Optional[IDTokenCredentials]=None, **context) -> bool:
-        data = self.read_gbq_base(from_query, from_pid, name, reauth, audience, account, credentials)
-        return self.upload_gspread(to_key, to_sheet, data, mode, cell, name=name, account=account, **context)
+    def update_data(self, from_key=str(), from_sheet=str(), from_query=str(), from_pid=str(),
+                    to_key=str(), to_sheet=str(), to_table=str(), to_pid=str(),
+                    mode: Literal["fail","replace","append","upsert"]="append", default=None, head=1, headers=None,
+                    str_cols: NumericiseIgnore=list(), to: Optional[Literal["desc","name"]]="desc",
+                    rename: Optional[RenameMap]=None, cell=str(), schema: Optional[BigQuerySchema]=None, progress=True,
+                    partition=str(), partition_by: Literal["auto","second","minute","hour","day","date"]="auto",name=str(),
+                    reauth=False, audience=str(), account: Account=dict(), credentials: Optional[IDTokenCredentials]=None,
+                    **context) -> bool:
+        if from_key and from_sheet:
+            data = self.read_gs_base(from_key, from_sheet, name, account, default, head, headers, str_cols, to, rename)
+        elif from_query and from_pid:
+            data = self.read_gbq_base(from_query, from_pid, name, reauth, audience, account, credentials)
+        else: return False
+        if to_key and to_sheet:
+            return self.upload_gspread(to_key, to_sheet, data, mode, cell, name=name, account=account, **context)
+        elif to_table and to_pid:
+            return self.upload_gbq(
+                to_table, to_pid, data, mode, schema, str(), progress, partition, partition_by,
+                name, reauth, audience, account, credentials, **context)
+        else: return False
 
 
 ###################################################################
@@ -428,38 +427,28 @@ def load_gspread(key: str, sheet: str, account: Account=dict()) -> Worksheet:
 
 def gs_loaded(func):
     @functools.wraps(func)
-    def wrapper(key=str(), sheet=str(), account: Account=dict(), *args, gs: Optional[Worksheet]=None, **kwargs):
+    def wrapper(key=str(), sheet=str(), *args, account: Account=dict(), gs: Optional[Worksheet]=None, **kwargs):
         if not gs:
             gs = load_gspread(key, sheet, account)
-        return func(*args, key=key, sheet=sheet, account=account, gs=gs, **kwargs)
+        return func(key, sheet, *args, account=account, gs=gs, **kwargs)
     return wrapper
 
 
-def _cast_boolean(__object) -> Union[Any,bool]:
+def _cast_boolean(__object) -> Union[bool,Any]:
     return {"TRUE":True, "FALSE":False}.get(__object, __object)
 
 
-def _to_excel_date(date: Datetime, default) -> int:
-    if not isinstance(date, dt.date): return default
+def _to_excel_date(__object) -> Union[int,Any]:
+    if isna(__object): return None
+    elif not isinstance(__object, dt.date): return __object
     offset = 693594
-    days = date.toordinal() - offset
-    if isinstance(date, dt.datetime):
-        seconds = (date.hour*60*60 + date.minute*60 + date.second)/(24*60*60)
+    days = __object.toordinal() - offset
+    if isinstance(__object, dt.datetime):
+        seconds = (__object.hour*60*60 + __object.minute*60 + __object.second)/(24*60*60)
         return days + seconds
     return days
 
 
-def _validate_gs_format(data: TabularData, action: Literal["read","update"]="read",
-                        fields: Optional[IndexLabel]=list(), default=None, if_null: Literal["drop","pass"]="pass",
-                        reorder=True, return_type: Optional[TypeHint]=None, rename: RenameMap=dict(),
-                        convert_first=False, rename_first=False, filter_first=False) -> TabularData:
-    if action == "read":
-        apply = lambda x: cast_datetime_format(x, default=_cast_boolean(x))
-    elif action == "update": apply = lambda x: _to_excel_date(x, default=x)
-    else: raise ValueError(INVALID_GS_ACTION_MSG(action))
-    return apply_data(all_keys=True, **drop_dict(locals(), "action", inplace=False))
-
-
 @gs_loaded
 def read_gspread(key: str, sheet: str, account: Account=dict(), gs: Optional[Worksheet]=None,
                 fields: Optional[IndexLabel]=list(), default=None, if_null: Literal["drop","pass"]="pass",
@@ -468,9 +457,10 @@ def read_gspread(key: str, sheet: str, account: Account=dict(), gs: Optional[Wor
                 convert_first=True, rename_first=True, filter_first=True) -> TabularData:
     if isinstance(numericise_ignore, bool): numericise_ignore = ["all"] if numericise_ignore else list()
     data = gs.get_all_records(head=head, default_blank=default, numericise_ignore=numericise_ignore, expected_headers=headers)
-    return _validate_gs_format(data, action="read",
-        fields=fields, default=default, if_null=if_null, reorder=reorder, return_type=return_type, rename=rename,
-        convert_first=convert_first, rename_first=rename_first, filter_first=filter_first)
+    apply = lambda x: cast_datetime_format(x, default=_cast_boolean(x))
+    return apply_data(data,
+        apply=apply, all_keys=True, fields=fields, default=default, if_null=if_null, reorder=reorder,
+        return_type=return_type, rename=rename, convert_first=convert_first, rename_first=rename_first, filter_first=filter_first)
 
 
 @gs_loaded
@@ -483,9 +473,9 @@ def clear_gspead(key: str, sheet: str, account: Account=dict(), gs: Optional[Wor
 @gs_loaded
 def update_gspread(key: str, sheet: str, data: TabularData, account: Account=dict(),
                     gs: Optional[Worksheet]=None, col='A', row=0, cell=str(), clear=False, clear_header=False):
-    records = _validate_gs_format(data, action="update", return_type="records")
+    records = to_records(data)
     if not records: return
-    values = [[value if pd.notna(value) else None for value in record.values()] for record in records]
+    values = [[_to_excel_date(__value) for __value in __m.values()] for __m in records]
     if clear: clear_gspead(gs=gs, include_header=clear_header)
     cell = cell if cell else (col+str(row if row else len(gs.get_all_records())+2))
     gs.update(cell, values)
@@ -525,7 +515,7 @@ def _validate_schema(data: pd.DataFrame, schema: Optional[BigQuerySchema]=None,
     context = {field["name"]:BIGQUERY_TYPE_CAST(field["type"], fillna) for field in schema}
     data = apply_df(cloc(data, list(context.keys()), if_null="pass"), **context)
     if df_exists(data, drop_na=True): return data
-    else: raise InvalidSchema(INVALID_SCHEMA_MSG, local_schema=data.dtypes.to_frame().to_dict()[0], remote_schema=schema)
+    else: raise InvalidSchema(SCHEMA_MISMATCH_MSG, local_schema=data.dtypes.to_frame().to_dict()[0], remote_schema=schema)
 
 
 @gbq_authorized
@@ -568,13 +558,12 @@ def to_gbq_partition(table: str, project_id: str, data: pd.DataFrame, reauth=Fal
             data[data[partition]==part].to_gbq(**context)
 
 
-def _validate_upsert_key(data: pd.DataFrame, base: pd.DataFrame, key=str(),
+def validate_primary_key(data: pd.DataFrame, base: pd.DataFrame, key=str(),
                         schema: Optional[BigQuerySchema]=None, index=0) -> str:
-    if not (key and isinstance(key, str)) and (schema and is_records(schema)):
-        keys = [field.get("name") for field in schema if field.get("mode") == "REQUIRED"]
-        key = iloc(keys, index, default=str())
+    if not key:
+        key = validate_gbq_schema(schema).get_primary_key(index=index)
     if key and (key in data) and (key in base): return key
-    else: raise ValueError(INVALID_UPSERT_KEY_MSG)
+    else: raise ValueError(INVALID_PRIMARY_KEY_MSG)
 
 
 @gbq_authorized
@@ -585,6 +574,6 @@ def upsert_gbq(table: str, project_id: str, data: pd.DataFrame, base: Optional[p
     if validate: data = _validate_schema(data, schema, fillna=fillna)
     context = dict(reauth=reauth, credentials=credentials)
     if df_empty(base): base = read_gbq(table, project_id, **context)
-    key = _validate_upsert_key(data, base, key, schema)
+    key = validate_primary_key(data, base, key, schema)
     data = data.set_index(key).combine_first(base.set_index(key)).reset_index()
     data.to_gbq(table, project_id, if_exists=if_exists, table_schema=schema, progress_bar=progress, **context)
