@@ -193,7 +193,7 @@ SET_TYPES = [set, Set, "set"]
 DICT_TYPES = [Dict, dict, "dict", "dictionary"]
 RECORDS_TYPES = [Records, "records"]
 DATAFRAME_TYPES = [DataFrame, "dataframe", "df", "pandas", "pd"]
-TAG_TYPES = [Tag, "source", "tag", "bs4"]
+TAG_TYPES = [Tag, "tag", "bs4", "html", "source"]
 
 TYPE_LIST = {
     bool: BOOLEAN_TYPES,
@@ -427,12 +427,12 @@ def get_annotation_name(annotation, custom_type=False) -> str:
     return annotation
 
 
-def get_annotation_typehint(annotation, custom_type=True) -> str:
+def get_annotation_typehint(annotation, custom_type=True) -> Union[str,List[str]]:
     annotation = get_annotation_name(annotation, custom_type=custom_type)
     return _decap(annotation, "Union", split=',')
 
 
-def get_annotation_type(annotation, custom_type=True) -> str:
+def get_annotation_type(annotation, custom_type=True) -> Union[str,List[str]]:
     name = get_annotation_typehint(annotation, custom_type=custom_type)
     if isinstance(name, List): return [__name for __name in map(get_annotation_type, name) if __name]
     name = name.replace("datetime.","").lower()
